use std::io::Write;

use crate::{constants::{MAGIC, TILE_SZ}, transform};


struct PixelTile {
    // i32: representation that supports Walsh-Hadamard
    pixels: [i32; TILE_SZ * TILE_SZ]
}
struct CoefTile {
    coefs: [i32; TILE_SZ * TILE_SZ]
}


pub fn compress(
    width: u16,
    height: u16,
    // n_components: u16,
    layers: &[&[u8]],
) -> Vec<u8> {
    let mut output = vec![];

    // validation
    for l in 0..layers.len() {
        assert!(layers[l].len() == width as usize * height as usize);
    }

    // write header
    output.write(&MAGIC).unwrap();
    output.write(&(width as u16).to_le_bytes()).unwrap();
    output.write(&(height as u16).to_le_bytes()).unwrap();
    // output.write(&n_components.to_le_bytes()).unwrap();
    output.write(&(layers.len() as u16).to_le_bytes()).unwrap();

    let mut tiles = vec![];
    for layer in layers.iter() {
        for x0 in (0..width).step_by(TILE_SZ) {
            for y0 in (0..height).step_by(TILE_SZ) {
                let pixel_tile = PixelTile::from_layer(
                    x0 as usize, y0 as usize, layer, 
                    width as usize, height as usize
                );
                let coef_tile = CoefTile::from_pixel_tile(&pixel_tile);
                tiles.push(coef_tile);
            }
        }
    }

    for t in tiles.iter() {
        t.write_zero(&mut output);
    }
    for t in tiles.iter() {
        t.write_rest(&mut output);
    }

    output
}

impl PixelTile {
    fn from_layer(
        x0: usize,
        y0: usize,
        layer: &[u8],
        width: usize,
        height: usize
    ) -> PixelTile {
        let mut pixels = [0; TILE_SZ * TILE_SZ];
        for x in 0..TILE_SZ {
            for y in 0..TILE_SZ {
                let src_x = x0 + x;
                let src_y = y0 + y;
                pixels[y * TILE_SZ + x] = 
                    if src_x < width && src_y < height { 
                        layer[src_y * width + src_x] as i32
                    } else {
                        0
                    };
            }
        }

        return PixelTile { pixels };
    }
}

impl CoefTile {
    fn from_pixel_tile(pt: &PixelTile) -> CoefTile {
        let mut coefs = pt.pixels.clone();

        // rows
        for y in 0..TILE_SZ {
            transform::encode(&mut coefs, y * TILE_SZ, 1);
        }

        // columns
        for x in 0..TILE_SZ {
            transform::encode(&mut coefs, x * TILE_SZ, 1);
        }

        return CoefTile { coefs }
    }

    fn write_zero(&self, output: &mut Vec<u8>) {
        output.write(&self.coefs[0].to_le_bytes()).unwrap();
    }

    fn write_rest(&self, output: &mut Vec<u8>) {
        for i in 1..self.coefs.len() {
            output.write(&self.coefs[i].to_le_bytes()).unwrap();
        }
    }
}