use std::io::{BufReader, Read};

use crate::{constants::{MAGIC, TILE_SZ}, transform};

struct PixelTile {
    // i32: representation that supports Walsh-Hadamard
    pixels: [i32; TILE_SZ * TILE_SZ]
}

#[derive(Clone)]
struct CoefTile {
    coefs: [i32; TILE_SZ * TILE_SZ]
}

pub fn decompress(
    input: &[u8]
) -> (u16, u16, Vec<Vec<u8>>) {
    let mut reader = BufReader::new(input);

    // read header
    let mut magic_buffer = [0; MAGIC.len()];
    reader.read_exact(&mut magic_buffer).unwrap();

    assert_eq!(MAGIC, magic_buffer);

    let mut u16_buffer = [0; 2];
    reader.read_exact(&mut u16_buffer).unwrap();
    let width = u16::from_le_bytes(u16_buffer);
    reader.read_exact(&mut u16_buffer).unwrap();
    let height = u16::from_le_bytes(u16_buffer);
    reader.read_exact(&mut u16_buffer).unwrap();
    let n_layers = u16::from_le_bytes(u16_buffer);

    let width_in_tiles = (width as usize + TILE_SZ - 1) / TILE_SZ;
    let height_in_tiles = (height as usize + TILE_SZ - 1) / TILE_SZ;

    let n_tiles = width_in_tiles * height_in_tiles * (n_layers as usize);
    let mut tiles = vec![CoefTile::new(); n_tiles];
    for i in 0..n_tiles {
        tiles[i].load_zero(&mut reader);
    }
    for i in 0..n_tiles {
        tiles[i].load_rest(&mut reader);
    }

    let mut tile_i = 0;
    let mut layers = vec![vec![0; width as usize * height as usize]; n_layers as usize];
    for layer in 0..n_layers {
        for x0 in (0..width).step_by(TILE_SZ) {
            for y0 in (0..height).step_by(TILE_SZ) {
                let pixel_tile = tiles[tile_i].to_pixel_tile();
                tile_i += 1;
                pixel_tile.to_layer(
                    x0 as usize, y0 as usize, &mut layers[layer as usize], 
                    width as usize, height as usize
                );
            }
        }
    }
    return (width, height, layers)
}

impl CoefTile {
    fn new() -> CoefTile {
        CoefTile { coefs: [0; TILE_SZ * TILE_SZ] }
    }
    
    fn load_zero(&mut self, reader: &mut BufReader<&[u8]>) {
        let mut coef_zero_buffer = [0; 4];
        reader.read_exact(&mut coef_zero_buffer).unwrap();
        self.coefs[0] = i32::from_le_bytes(coef_zero_buffer);
    }

    fn load_rest(&mut self, reader: &mut BufReader<&[u8]>) {
        let mut coef_buffer = [0; 4];
        for i in 1..self.coefs.len() {
            reader.read_exact(&mut coef_buffer).unwrap();
            self.coefs[i] = i32::from_le_bytes(coef_buffer);
        }
    }
    
    fn to_pixel_tile(&self) -> PixelTile {
        let mut pixels = self.coefs.clone();

        // columns
        for x in 0..TILE_SZ {
            transform::decode(&mut pixels, x * TILE_SZ, 1);
        }

        // rows
        for y in 0..TILE_SZ {
            transform::decode(&mut pixels, y * TILE_SZ, 1);
        }

        return PixelTile { pixels };
    }
}

impl PixelTile {
    fn to_layer(&self, x0: usize, y0: usize, layer: &mut [u8], width: usize, height: usize) {
        for x in 0..TILE_SZ {
            for y in 0..TILE_SZ {
                let dst_x = x0 + x;
                let dst_y = y0 + y;
                if dst_x < width && dst_y < height { 
                    layer[dst_y * width + dst_x] = self.pixels[y * TILE_SZ + x] as u8;
                }
            }
        }
    }
}